# vim: set fileencoding=utf-8 :

from django.db import models
from django.contrib.auth.models import User
from django.conf import settings

from django.utils.translation import ugettext_lazy as _

from decimal import Decimal
from datetime import datetime,date,timedelta
from dateutil.relativedelta import relativedelta

the_app_label = 'katrin'

YES_NO = (
	(1, _('Yes')),
	(0, _('No')),
)

class FilterGroup(models.Model):
	"""Name is needed for usage operation description
	"""
	name = models.CharField(_('Name'), max_length=80)
	def __unicode__(self):
		return u"%s" % (self.name)
	class Meta:
		db_table = u'filter_groups'
		verbose_name = _('filter group')
		verbose_name_plural = _('Filter groups')
		app_label = the_app_label

class Period:
	Divisibles = {
		'Hour':		1,
		'Day':		2,
		'Month':	3,
	}

	@classmethod
	def get_period_start(cls, date_time, divisible, shifts=0):
		if divisible not in cls.Divisibles.values():
			raise ValueError('Unknown divisible %s. Use one of %s' % (divisible, cls.Divisibles.values()))
		return {
				cls.Divisibles['Hour']: lambda dt: datetime(year=dt.year, month=dt.month, day=dt.day, hour=dt.hour)+timedelta(hours=shifts), 
				cls.Divisibles['Day']: lambda dt: datetime(year=dt.year, month=dt.month, day=dt.day)+timedelta(days=shifts), 
				cls.Divisibles['Month']: lambda dt: datetime(year=dt.year, month=dt.month, day=dt.day)+relativedelta(months=shifts)
				}[divisible](date_time)

	@classmethod
	def get_period_end(cls, date_time, divisible):
		return cls.get_period_start(date_time, divisible, 1)

class BillingPeriod(models.Model):
	""" You can create own Billing Periods. You should provide period divisible to define which 
	statistics agregation level to use.
	"""
	DIVISIBLES_CHOICES = (
			(Period.Divisibles['Hour'], _('Hour')),
			(Period.Divisibles['Day'], _('Day')),
			(Period.Divisibles['Month'], _('Month')),
	)
	duration = models.CharField(_('Duration'), max_length=20, default=_('Month'))
	divisible = models.IntegerField(_('Divisible'), default=Period.Divisibles['Hour'], choices=DIVISIBLES_CHOICES)
	class Meta:
		db_table = u'billing_periods'
		verbose_name = _('billing period')
		verbose_name_plural = _('Billing periods')
		app_label = the_app_label

	def __unicode__(self):
		return "%s" % self.duration

OperationType = {
		'Debit':		1,
		'Credit':		2,
		'Activation':	3,
		'Usage':		4,
		'Periodic':		5,
		'Refund':		6,
		'Closure':		7,
}

OperationStatus = {
		'Closed':	0,
		'No bill':	1,
		}

class Operation(models.Model):
	"""Operation has a type. Billing_period doesn't make sense for Debit and Credit operations"""
	TYPE_CHOICES = (
			(OperationType['Debit'], _('Debit')),
			(OperationType['Credit'], _('Credit')),
			(OperationType['Activation'], _('Activation fee')),
			(OperationType['Usage'], _('Usage fee')),
			(OperationType['Periodic'], _('Periodic fee')),
			(OperationType['Refund'], _('Refund')),
			(OperationType['Closure'], _('Billing period closure')),
	)
	STATUS_CHOICES = (
			(OperationStatus['Closed'], _('Closed')),
			(OperationStatus['No bill'], _('No bill')),
	)
	description = models.CharField(_('Description'), default=_("No description"), max_length=200)
	account = models.ForeignKey('Account', verbose_name=_('Account'), related_name='operations')
	type = models.IntegerField(_('Type'), choices=TYPE_CHOICES)
	service = models.ForeignKey('Service', verbose_name=_('Service'), related_name='operations', null=True, blank=True)
	debit = models.DecimalField(_('Debit'), max_digits=10, decimal_places=4, default=Decimal('0'))
	credit = models.DecimalField(_('Credit'), max_digits=10, decimal_places=4, default=Decimal('0'))
	date_time = models.DateTimeField(_('Date and time'), default=datetime.now())
	status = models.IntegerField(_('Status'), choices=STATUS_CHOICES, default=OperationStatus['Closed'],)

	class Meta:
		db_table = u'operations'
		verbose_name = _('operation')
		verbose_name_plural = _('Operations')
		app_label = the_app_label

	@classmethod
	def do_activation(cls, userservice):
		if userservice.tariff.activation_fee > 0:
			desc = u"%s %s %s" % (_('Activation fee'), _('according to tariff'), userservice.tariff)
			op = Operation(
					description = desc,
					type = OperationType['Activation'],
					account = userservice.account,
					debit = userservice.tariff.activation_fee,
					service = userservice.service,
					)
			op.init_status(userservice)
			op.save()
			return op

	@classmethod
	def do_periodic(cls, userservice):
		if userservice.tariff.periodic_fee > 0:
			desc = u"%s %s %s" % (_('Periodic fee'), _('according to tariff'), userservice.tariff)
			op = Operation(
					description = desc,
					type = OperationType['Periodic'],
					account = userservice.account,
					debit = userservice.tariff.periodic_fee,
					service = userservice.service,
					)
			op.init_status(userservice)
			op.save()
			return op

	@classmethod
	def do_usage(cls, userservice, description, debit, change_balance=False):
		"""By default it is supposed that katrin kernel has already changed balance"""
		if debit < 0:
			raise ValueError("Debit can't be negative")

		if debit > 0:
			op = Operation(
					description = description,
					type = OperationType['Usage'],
					account = userservice.account,
					debit = debit,
					service = userservice.service,
					)
			op.init_status(userservice)
			op.save(change_balance)
			return op

	@classmethod
	def do_closure(cls, userservice):
		desc = u"%s %s %s" % (_('Billing period closure'), _('for service'), userservice.service)
		op = Operation(
				description = desc,
				type = OperationType['Closure'],
				account = userservice.account,
				service = userservice.service,
				status = OperationStatus['Closed'],
				)
		op.save()
		return op

	@classmethod
	def get_first_debit_operation(cls, userservice):
		op = cls.objects.filter(service=userservice.service, account=userservice.account, debit__gt=0)[:1]
		if op:
			return op[0]
		else:
			return None
	
	@classmethod
	def get_last_closure(cls):
		try:
			return cls.objects.filter(type=OperationType['Closure']).latest('date_time')
		except cls.DoesNotExist:
			return None

# Instance methods
	def __unicode__(self):
		return self.description

	def get_service(self):
		if self.service:
			return self.service
		return u"%s" % _('No')
	get_service.short_description = _('Service')

	def init_status(self, userservice):
		if userservice.account.postpaid:
			self.status = OperationStatus['No bill']
		else:
			self.status = OperationStatus['Closed']

	def save(self, change_balance=True):
		# If usage is charging by katrin kernel in real time
		# then set change_balance=False
		delta = 0
		if self.type in (
				OperationType['Activation'], 
				OperationType['Periodic'],
				OperationType['Refund'], 
				OperationType['Usage'],
				OperationType['Closure'],
				) and not self.service:
			raise ValueError(u"Service shouldn't be empty for operation type %s" % self.type)

		if self.type in (
				OperationType['Debit'], 
				OperationType['Activation'], 
				OperationType['Periodic'],
				):
			self.credit=0
			delta = -self.debit
		elif self.type in (
				OperationType['Usage'],
				):
			self.credit=0
			delta = -self.debit
		elif self.type in (
				OperationType['Credit'], 
				OperationType['Refund'], 
				):
			self.debit=0
			delta = self.credit
		elif self.type == OperationType['Closure']:
			if not self.account:
				raise ValueError("Operation with closure type must have service")
			self.debit=0
			self.credit=0
		if change_balance:
			self.account.balance += delta
		save_account = True
		# Don't save account if operation was edited
		if self.id:
			save_account = False
		super(Operation, self).save()
		if save_account:
			self.account.save(None, self)

	def delete(self):
		self.account_history.operation=None
		self.account_history.save()
		super(Operation, self).delete()

class AccountHistory(models.Model):
	date_time = models.DateTimeField(_('Date and time'), default=datetime.now())
	admin = models.ForeignKey(User, verbose_name=_('Admin'), null=True, blank=True)
	account = models.ForeignKey('Account', verbose_name=_('Account'))
	balance = models.DecimalField(_('Balance'), max_digits=10, decimal_places=4, default=0.0)
	credit = models.DecimalField(_('Credit'), max_digits=10, decimal_places=4, default=0.0)
	blocked = models.IntegerField(_('Blocked'), default=0, choices=YES_NO)
	operation = models.OneToOneField(Operation, null=True, blank=True, verbose_name=_('Operation'), related_name="account_history")

	class Meta:
		db_table = u'accounts_history'
		verbose_name = _('account history')
		verbose_name_plural = _('History of accounts')
		app_label = the_app_label

	def __unicode__(self):
		return u"%s (%s %s)" % (self.account, self.admin, self.date_time)

	def get_admin(self):
		if self.admin:
			return self.admin
		else:
			return u"Katrin system"
	get_admin.short_description = _("Admin")

	def get_operation(self):
		if self.operation:
			return self.operation
		else:
			return _('No')
	get_operation.short_description = _("Operation")

class Bill(models.Model):
	""" Manages prepared bills. At the moment only bills for postpaid accounts are supported. Bills can be exported
	to external system which can handle prepared bills and, for example, make billing statement.
	"""
	bill_to = models.CharField(_('Bill to'), max_length=50)
	bill_to_extid = models.CharField(_('External id of bill recipient'), blank=True, null=True, max_length=30)
	date = models.DateField(_('Date'), default = date.today())

	class Meta:
		db_table = u'bills'
		verbose_name = _('bill')
		verbose_name_plural = _('Bills')
		app_label = the_app_label
	
	def subtotal(self):
		subtotal = 0
		for entry in self.entries.all():
			subtotal += entry.subtotal()
		return subtotal

	def vat(self):
		vat = 0
		for entry in self.entries.all():
			vat += entry.vat_amount
		return vat

	def total(self):
		total = 0
		for entry in self.entries.all():
			total += entry.total()
		return total
	total.short_description = _('Total')
	
	def __unicode__(self):
		return u'%i' % self.pk

class BillEntry(models.Model):
	""" Bill entries are part of bill. VAT stands for Value Added Tax. In Russia it is
	called NDS. Each bill entry corresponds to some operation
	"""
	class Meta:
		db_table = u'bill_entries'
		verbose_name = _('bill entry')
		verbose_name_plural = _('Bill entries')
		app_label = the_app_label

	bill = models.ForeignKey(Bill, verbose_name=_('Bill'), related_name="entries")
	operation = models.OneToOneField(Operation, verbose_name=_('Operation'))
	net_amount = models.DecimalField(_('Net amount'), max_digits=10, decimal_places=4, default=Decimal('0'))
	discount = models.DecimalField(_('Discount'), max_digits=10, decimal_places=4, default=Decimal('0'))
	vat_amount = models.DecimalField(_('VAT amount'), max_digits=10, decimal_places=4, default=Decimal('0'))
	
	def __unicode__(self):
		return u'%i' % self.pk

	def subtotal(self):
		return self.net_amount - self.discount

	def total(self):
		return self.subtotal() + self.vat_amount
	total.short_description = _('Entry total')


class RefillLog(models.Model):
	id = models.AutoField(primary_key=True)
	datetime = models.DateTimeField(default=datetime.now())
	admin = models.ForeignKey(User, verbose_name=_('Admin'))
	account = models.ForeignKey('Account')
	refill_size = models.DecimalField(_('Increase size'), max_digits=5, decimal_places=2)
	class Meta:
		verbose_name = _('increase balance')
		verbose_name_plural = _('History of increases')
		app_label = the_app_label


# This classes only for show links in admin main page.
class Refill(models.Model):
	id = models.AutoField(primary_key=True)
	class Meta:
		db_table = u'not_table'
		verbose_name_plural = _('Increase balance')
		app_label = the_app_label
